package org.xbib.event.async.io;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.AsynchronousFileChannel;
import java.nio.channels.CompletionHandler;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.concurrent.CompletableFuture;

import static java.nio.ByteBuffer.wrap;
import static java.nio.channels.AsynchronousFileChannel.open;

/**
 * Asynchronous non-blocking write operations with a {@code CompletableFuture} based API.
 * These operations use an underlying {@code AsynchronousFileChannel}.
 * All methods are asynchronous including the {@code close()} which chains a continuation
 * on last resulting write {@code CompletableFuture} to close the {@code AsyncFileChannel} on completion.
 */
public class AsyncFileWriter implements AutoCloseable {

    final AsynchronousFileChannel asyncFile;
    /**
     * File position after last write operation completion.
     */
    private CompletableFuture<Integer> currentPosition = CompletableFuture.completedFuture(0);

    public AsyncFileWriter(AsynchronousFileChannel asyncFile) {
        this.asyncFile = asyncFile;
    }

    public AsyncFileWriter(Path file, StandardOpenOption... options) throws IOException {
        this(open(file, options));
    }

    static CompletableFuture<Integer> write(
            AsynchronousFileChannel asyncFile,
            ByteBuffer buf,
            int position) {
        CompletableFuture<Integer> promise = new CompletableFuture<>();
        asyncFile.write(buf, position, null, new CompletionHandler<Integer, Object>() {
            @Override
            public void completed(Integer result, Object attachment) {
                promise.complete(result);
            }

            @Override
            public void failed(Throwable exc, Object attachment) {
                promise.completeExceptionally(exc);
            }
        });
        return promise;
    }

    public CompletableFuture<Integer> getPosition() {
        return currentPosition;
    }

    /**
     * Writes the given String appended with a newline separator
     * and returns a CompletableFuture of the final file index
     * after the completion of the corresponding write operation.
     */
    public CompletableFuture<Integer> writeLine(CharSequence str) {
        return write(str + System.lineSeparator());
    }

    /**
     * Writes the given String and returns a CompletableFuture of
     * the final file index after the completion of the corresponding
     * write operation.
     */
    public CompletableFuture<Integer> write(String str) {
        return write(str.getBytes());
    }

    /**
     * Writes the given byte array and returns a CompletableFuture of
     * the final file index after the completion of the corresponding
     * write operation.
     */
    public CompletableFuture<Integer> write(byte[] bytes) {
        return write(wrap(bytes));
    }

    /**
     * Writes the given byte buffer and returns a CompletableFuture of
     * the final file index after the completion of the corresponding
     * write operation.
     */
    public CompletableFuture<Integer> write(ByteBuffer bytes) {
        /*
         * Wee need to update currentPosition field to keep track.
         * The currentPosition field is used on close() method, which chains
         * a continuation to close the AsyncFileChannel.
         */
        currentPosition = currentPosition.thenCompose(index -> {
            CompletableFuture<Integer> size = write(asyncFile, bytes, index);
            return size.thenApply(length -> length + index);
        });
        return currentPosition;
    }

    /**
     * Asynchronous close operation.
     * Chains a continuation on CompletableFuture resulting from last write operation,
     * which closes the AsyncFileChannel on completion.
     *
     * @throws IOException
     */
    @Override
    public void close() throws IOException {
        if (asyncFile != null) {
            currentPosition.whenComplete((res, ex) ->
                    AsyncFiles.closeAfc(asyncFile)
            );
        }
    }
}
