package org.xbib.event.bus.util;

import java.lang.reflect.Array;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.lang.reflect.WildcardType;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

public abstract class TypeToken<T> {

    private final Type runtimeType;

    private TypeToken(Type type) {
        this.runtimeType = type;
    }

    public static <T> TypeToken<T> of(Class<T> type) {
        return new SimpleTypeToken<>(type);
    }

    public static TypeToken<?> of(Type type) {
        return new SimpleTypeToken<>(type);
    }

    public final TypeSet getTypes() {
        return new TypeSet();
    }

    public final Class<? super T> getRawType() {
        return getRawTypes().iterator().next();
    }

    private static final class SimpleTypeToken<T> extends TypeToken<T> {

        SimpleTypeToken(Type type) {
            super(type);
        }
    }

    private Collection<Class<? super T>> getRawTypes() {
        final List<Class<?>> builder = new ArrayList<>();
        new TypeVisitor() {
            @Override
            void visitTypeVariable(TypeVariable<?> t) {
                visit(t.getBounds());
            }

            @Override
            void visitWildcardType(WildcardType t) {
                visit(t.getUpperBounds());
            }

            @Override
            void visitParameterizedType(ParameterizedType t) {
                builder.add((Class<?>) t.getRawType());
            }

            @Override
            void visitClass(Class<?> t) {
                builder.add(t);
            }

            @Override
            void visitGenericArrayType(GenericArrayType t) {
                builder.add(getArrayClass(of(t.getGenericComponentType()).getRawType()));
            }
        }.visit(runtimeType);
        @SuppressWarnings({"unchecked", "rawtypes"})
        Collection<Class<? super T>> result = (Collection) builder;
        return result;
    }

    private abstract static class TypeCollector<K> {

        static final TypeCollector<Class<?>> FOR_RAW_TYPE =
                new TypeCollector<>() {
                    @Override
                    Class<?> getRawType(Class<?> type) {
                        return type;
                    }

                    @Override
                    Iterable<? extends Class<?>> getInterfaces(Class<?> type) {
                        return Arrays.asList(type.getInterfaces());
                    }

                    @Override
                    Class<?> getSuperclass(Class<?> type) {
                        return type.getSuperclass();
                    }
                };


        Collection<K> collectTypes(Iterable<? extends K> types) {
            Map<K, Integer> map = new HashMap<>();
            for (K type : types) {
                collectTypes(type, map);
            }
            @SuppressWarnings("unchecked")
            Comparator<Integer> comparator = (Comparator<Integer>) Comparator.naturalOrder().reversed();
            return sortKeysByValue(map, comparator);
        }

        private int collectTypes(K type, Map<? super K, Integer> map) {
            Integer existing = map.get(type);
            if (existing != null) {
                return existing;
            }
            int aboveMe = getRawType(type).isInterface() ? 1 : 0;
            for (K interfaceType : getInterfaces(type)) {
                aboveMe = Math.max(aboveMe, collectTypes(interfaceType, map));
            }
            K superclass = getSuperclass(type);
            if (superclass != null) {
                aboveMe = Math.max(aboveMe, collectTypes(superclass, map));
            }
            map.put(type, aboveMe + 1);
            return aboveMe + 1;
        }

        private static <K, V> Collection<K> sortKeysByValue(Map<K, V> map,
                                                            Comparator<? super V> valueComparator) {
            Comparator<K> keyOrdering =
                    (left, right) -> valueComparator.compare(map.get(left), map.get(right));
            List<K> keys = new ArrayList<>(map.keySet());
            keys.sort(keyOrdering);
            return keys;
        }

        abstract Class<?> getRawType(K type);

        abstract Iterable<? extends K> getInterfaces(K type);

        abstract K getSuperclass(K type);

    }

    public class TypeSet  {

        TypeSet() {}

        public Set<Class<? super T>> rawTypes() {
            @SuppressWarnings({"unchecked", "rawtypes"})
            Collection<Class<? super T>> collectedTypes =
                    (Collection) TypeCollector.FOR_RAW_TYPE.collectTypes(getRawTypes());
            return new HashSet<>(collectedTypes);
        }
    }

    static Class<?> getArrayClass(Class<?> componentType) {
        return Array.newInstance(componentType, 0).getClass();
    }
}
