package org.xbib.event.log;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.Objects;
import java.util.concurrent.ScheduledExecutorService;

public class RotationConfig {

    private final Path path;

    private final RotatingFilePattern filePattern;

    private final ScheduledExecutorService executorService;

    private final Collection<RotationPolicy> policies;

    private final boolean append;

    private final boolean compress;

    private final Clock clock;

    private final RotationListener rotationListener;

    private RotationConfig(Builder builder) {
        this.path = builder.path;
        this.filePattern = builder.filePattern;
        this.executorService = builder.executorService;
        this.policies = builder.policies;
        this.append = builder.append;
        this.compress = builder.compress;
        this.clock = builder.clock;
        this.rotationListener = builder.listener;
    }

    public Path getPath() {
        return path;
    }

    public RotatingFilePattern getFilePattern() {
        return filePattern;
    }

    public Collection<RotationPolicy> getPolicies() {
        return policies;
    }

    public boolean isAppend() {
        return append;
    }

    public boolean isCompress() {
        return compress;
    }

    public Clock getClock() {
        return clock;
    }

    public RotationListener getListener() {
        return rotationListener;
    }

    public ScheduledExecutorService getScheduler() {
        return executorService;
    }

    @Override
    public boolean equals(Object instance) {
        if (this == instance) return true;
        if (instance == null || getClass() != instance.getClass()) return false;
        RotationConfig that = (RotationConfig) instance;
        return append == that.append &&
                compress == that.compress &&
                Objects.equals(path, that.path) &&
                Objects.equals(filePattern, that.filePattern) &&
                Objects.equals(executorService, that.executorService) &&
                Objects.equals(policies, that.policies) &&
                Objects.equals(clock, that.clock) &&
                Objects.equals(rotationListener, that.rotationListener);
    }

    @Override
    public int hashCode() {
        return Objects.hash(path, filePattern, executorService, policies, append, compress, clock, rotationListener);
    }

    @Override
    public String toString() {
        return String.format("RotationConfig{path=%s}", path);
    }

    public static Builder builder() {
        return new Builder();
    }

    public static class Builder {

        private Path path;

        private RotatingFilePattern filePattern;

        private ScheduledExecutorService executorService;

        private Collection<RotationPolicy> policies;

        private boolean append;

        private boolean compress;

        private Clock clock;

        private RotationListener listener;

        private Builder() {
            this.append = true;
            this.clock = SystemClock.getInstance();
            this.listener = LoggingRotationListener.getInstance();
        }

        public Builder path(Path path) {
            this.path = path;
            return this;
        }

        public Builder path(String fileName) {
            this.path = Paths.get(fileName);
            return this;
        }

        public Builder filePattern(RotatingFilePattern filePattern) {
            this.filePattern = filePattern;
            return this;
        }

        public Builder filePattern(String filePattern) {
            this.filePattern = RotatingFilePattern.builder().pattern(filePattern).build();
            return this;
        }

        public Builder executorService(ScheduledExecutorService executorService) {
            this.executorService = executorService;
            return this;
        }

        public Builder policies(Collection<RotationPolicy> policies) {
            this.policies = policies;
            return this;
        }

        public Builder policy(RotationPolicy policy) {
            if (policies == null) {
                policies = new LinkedHashSet<>();
            }
            policies.add(policy);
            return this;
        }

        public Builder append(boolean append) {
            this.append = append;
            return this;
        }

        public Builder compress(boolean compress) {
            this.compress = compress;
            return this;
        }

        public Builder clock(Clock clock) {
            this.clock = clock;
            return this;
        }

        public Builder listener(RotationListener listener) {
            this.listener = listener;
            return this;
        }

        public RotationConfig build() {
            validate();
            return new RotationConfig(this);
        }

        private void validate() {
            Objects.requireNonNull(path, "path");
            Objects.requireNonNull(filePattern, "filePattern");
            if (policies == null || policies.isEmpty()) {
                throw new IllegalArgumentException("empty policies");
            }
            Objects.requireNonNull(clock, "clock");
            Objects.requireNonNull(listener, "callback");
        }
    }
}
