package org.xbib.event.timer;

import org.xbib.event.Event;
import org.xbib.event.common.PayloadImpl;
import org.xbib.event.common.AbstractEventManagerService;
import org.xbib.event.common.EventManager;
import org.xbib.event.common.EventManagerService;
import org.xbib.event.persistence.FilePersistenceStore;
import org.xbib.event.persistence.PersistenceStore;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;

import java.io.IOException;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class TimerEventManagerService extends AbstractEventManagerService implements EventManagerService {

    private static final Logger logger = Logger.getLogger(TimerEventManagerService.class.getName());

    private EventManager eventManager;

    private final Map<String, TimerEventService> services;

    public TimerEventManagerService() {
        super();
        this.services = new LinkedHashMap<>();
    }

    @Override
    public TimerEventManagerService init(EventManager eventManager) {
        this.eventManager = eventManager;
        for (Map.Entry<String, Settings> entry : eventManager.getSettings().getGroups("event.timer").entrySet()) {
            String name = entry.getKey();
            Settings timerSettings = entry.getValue();
            try {
                PersistenceStore<String, Object> persistenceStore = new FilePersistenceStore(timerSettings, name);
                TimerEventService timerEventService = new TimerEventService(this, name, ZoneId.systemDefault(), persistenceStore);
                services.put(name, timerEventService);
                logger.log(Level.INFO, "timer " + name + " active: " + timerEventService);
            } catch (Exception e) {
                logger.log(Level.WARNING, "unable to activate timer " + name + ", reason: " + e.getMessage(), e);
            }
        }
        logger.log(Level.INFO, "after init: event services = " + services);
        return this;
    }

    @Override
    public void shutdown() throws IOException {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "closing timer " + entry.getKey());
            entry.getValue().close();
        }
    }

    public void publish(Event event) {
        eventManager.publish(event);
    }

    public boolean schedule(String name,
                            String timeSpec,
                            PayloadImpl payload) throws ParseException, IOException {
        if (services.containsKey(name)) {
            Span span = Chronic.parse(timeSpec);
            if (span != null) {
                ZonedDateTime zonedDateTime = span.getBeginCalendar();
                services.get(name).schedule(zonedDateTime.toInstant(), payload);
                logger.log(Level.INFO, "scheduled to " + zonedDateTime);
            } else {
                logger.log(Level.INFO, "timer event key " + name + ": can not understand time spec " + timeSpec);
                return false;
            }
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + name);
        }
        return false;
    }

    public boolean schedule(String service,
                            Instant instant,
                            PayloadImpl payload) throws IOException {
        if (services.containsKey(service)) {
            services.get(service).schedule(instant, payload);
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + service);
        }
        return false;
    }

    public void purge() {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "purging timer " + entry.getKey());
            entry.getValue().purge();
        }
    }
}
