package org.xbib.event.loop.thread;

import org.xbib.event.loop.EventExecutor;

import java.util.Objects;
import java.util.concurrent.Executor;
import java.util.concurrent.ThreadFactory;

/**
 * Allow to retrieve the {@link EventExecutor} for the calling {@link Thread}.
 */
public final class ThreadExecutorMap {

    private static final FastThreadLocal<EventExecutor> mappings = new FastThreadLocal<>();

    private ThreadExecutorMap() { }

    /**
     * Returns the current {@link EventExecutor} that uses the {@link Thread}, or {@code null} if none / unknown.
     */
    public static EventExecutor currentExecutor() {
        return mappings.get();
    }

    /**
     * Set the current {@link EventExecutor} that is used by the {@link Thread}.
     */
    private static void setCurrentEventExecutor(EventExecutor executor) {
        mappings.set(executor);
    }

    /**
     * Decorate the given {@link Executor} and ensure {@link #currentExecutor()} will return {@code eventExecutor}
     * when called from within the {@link Runnable} during execution.
     */
    public static Executor applyToExecutor(final Executor executor, final EventExecutor eventExecutor) {
        Objects.requireNonNull(executor, "executor");
        Objects.requireNonNull(eventExecutor, "eventExecutor");
        return command -> executor.execute(applyToRunnable(command, eventExecutor));
    }

    /**
     * Decorate the given {@link ThreadFactory} and ensure {@link #currentExecutor()} will return {@code eventExecutor}
     * when called from within the {@link Runnable} during execution.
     */
    public static ThreadFactory applyToThreadFactory(final ThreadFactory threadFactory, final EventExecutor eventExecutor) {
        Objects.requireNonNull(threadFactory, "command");
        Objects.requireNonNull(eventExecutor, "eventExecutor");
        return r -> threadFactory.newThread(applyToRunnable(r, eventExecutor));
    }

    /**
     * Decorate the given {@link Runnable} and ensure {@link #currentExecutor()} will return {@code eventExecutor}
     * when called from within the {@link Runnable} during execution.
     */
    public static Runnable applyToRunnable(final Runnable command, final EventExecutor eventExecutor) {
        Objects.requireNonNull(command, "command");
        Objects.requireNonNull(eventExecutor, "eventExecutor");
        return () -> {
            setCurrentEventExecutor(eventExecutor);
            try {
                command.run();
            } finally {
                setCurrentEventExecutor(null);
            }
        };
    }
}
