package org.xbib.event.syslog;

import io.netty.buffer.ByteBuf;
import io.netty.channel.ChannelHandlerContext;
import io.netty.handler.codec.MessageToMessageEncoder;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.time.format.DateTimeFormatter;
import java.util.List;
import java.util.Map;

public class MessageEncoder extends MessageToMessageEncoder<Message> {

    final DateTimeFormatter cefDateFormat;
    final Charset charset;
    final byte[] cef;
    final byte[] pipe;

    public MessageEncoder(DateTimeFormatter cefDateFormat) {
        this.cefDateFormat = cefDateFormat;
        this.charset = StandardCharsets.UTF_8;
        this.cef = "CEF:0".getBytes(charset);
        this.pipe = "|".getBytes(charset);
    }

    @Override
    protected void encode(ChannelHandlerContext context, Message message, List<Object> list) {
        switch (message.type()) {
            case CEF:
                encodeCEF(context, message, list);
                break;
            case RFC3164:
                encodeRFC3164(context, message, list);
                break;
            case RFC5424:
                encodeRFC5424(context, message, list);
                break;
            default:
                break;
        }
    }

    private void encodeCEF(ChannelHandlerContext context, Message message, List<Object> output) {
        final ByteBuf buffer = context.alloc().buffer();
        EncoderHelper.appendPriority(buffer, message);
        buffer.writeCharSequence(cefDateFormat.format(message.date()), charset);
        buffer.writeBytes(EncoderHelper.SPACE);
        buffer.writeCharSequence(message.host(), charset);
        buffer.writeBytes(EncoderHelper.SPACE);
        buffer.writeBytes(cef);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.deviceVendor(), charset);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.deviceProduct(), charset);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.deviceVersion(), charset);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.deviceEventClassId(), charset);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.name(), charset);
        buffer.writeBytes(pipe);
        buffer.writeCharSequence(message.severity(), charset);
        buffer.writeBytes(pipe);
        int index = 0;
        /*for (Map.Entry<String, Object> kvp : message.getMap().entrySet()) {
            if (index > 0) {
                buffer.writeBytes(EncoderHelper.SPACE);
            }
            buffer.writeCharSequence(kvp.getKey(), charset);
            buffer.writeBytes(EncoderHelper.EQUALS);
            buffer.writeCharSequence(kvp.getValue().toString(), charset);
            index++;
        }*/
        output.add(buffer);
    }

    private void encodeRFC3164(ChannelHandlerContext context, Message message, List<Object> output) {
        ByteBuf buffer = context.alloc().buffer();
        EncoderHelper.appendPriority(buffer, message);
        buffer.writeCharSequence(message.date().format(cefDateFormat), charset);
        buffer.writeCharSequence(" ", charset);
        buffer.writeCharSequence(message.host(), charset);
        buffer.writeCharSequence(" ", charset);
        buffer.writeCharSequence(message.tag(), charset);
        if (message.processId() != null) {
            buffer.writeCharSequence("[", charset);
            buffer.writeCharSequence(message.processId(), charset);
            buffer.writeCharSequence("]", charset);
        }
        buffer.writeCharSequence(": ", charset);
        buffer.writeCharSequence(message.message(), charset);
        output.add(buffer);
    }

    private void encodeRFC5424(ChannelHandlerContext context, Message message, List<Object> output) {
        final ByteBuf buffer = context.alloc().buffer();
        EncoderHelper.appendPriority(buffer, message);
        if (message.version() != null) {
            buffer.writeCharSequence(message.version().toString(), charset);
        }
        buffer.writeBytes(EncoderHelper.SPACE);
        buffer.writeCharSequence(message.date().format(cefDateFormat), charset);
        buffer.writeCharSequence(" ", charset);
        buffer.writeCharSequence(message.host(), charset);
        buffer.writeCharSequence(" ", charset);
        if (message.appName() != null) {
            buffer.writeCharSequence(message.appName(), charset);
        }
        if (message.processId() != null) {
            buffer.writeCharSequence(message.processId(), charset);
        } else {
            buffer.writeCharSequence(" -", charset);
        }
        if (message.messageId() != null) {
            buffer.writeCharSequence(message.messageId(), charset);
        } else {
            buffer.writeCharSequence(" -", charset);
        }
        buffer.writeCharSequence(" - ", charset);
        buffer.writeCharSequence(message.message(), charset);
        output.add(buffer);
    }
}
